==> kite-1.0b6.tar.gz: kite-1.0b6/modules/__internal/_cinvoke_callback.c <==
/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "kite_object.h"
#include "vm/kite_vm.h"
#include "config.h"

#ifdef BUILD_CINVOKE
#include "cinvoke/cinvoke.h"

/*****************************************************************************
 * Arrange to call given Kite method.
 ****************************************************************************/
void
CInvokeCallback_callIntoKite(CInvFunction *f, void *parameters[], 
                             void *returnout, void *userdata)
{
    char *param_types_string, *ret_type_string;
    kite_object_t *args = (kite_object_t*)userdata;
    kite_object_t *retString, *paramString, *userData, *method;
    kite_object_t *methodArgs;
    kite_object_t *threadInt;
    kite_object_t *tmp;
    kite_thread_t *thd;
    int i = 0;
    int curstack;
    
    (void)f;
    
    KITE_GET_METHOD_ARGUMENT(retString, 1);
    KITE_GET_METHOD_ARGUMENT(paramString, 2);
    KITE_GET_METHOD_ARGUMENT(userData, 3);
    KITE_GET_METHOD_ARGUMENT(method, 4);
    KITE_GET_METHOD_ARGUMENT(threadInt, 5);
    
    thd = (kite_thread_t*)KITE_GET_INTEGER(threadInt);
    param_types_string = KITE_GET_STRING_VALUE(paramString);
    methodArgs = kite_new_list(thd);
    kite_append_list(thd, methodArgs, userData);
    while(*param_types_string)
    {        
        switch(*param_types_string)
        {
            case 'c':
            {
                tmp = kite_new_integer(thd, *(char*)parameters[i]);
                break;
            }
            case 's':
            {
                tmp = kite_new_integer(thd, *(short*)parameters[i]);
                break;
            }
            case 'i':
            {
                tmp = kite_new_integer(thd, *(int*)parameters[i]);
                break;
            }
            case 'l':
            {
                tmp = kite_new_integer(thd, *(long*)parameters[i]);
                break;
            }
            case 'e':
            {
                tmp = kite_new_integer(thd, *(long long*)parameters[i]);
                break;
            }
            case 'f':
            {
                tmp = kite_new_float(thd, *(float*)parameters[i]);
                break;
            }
            case 'd':
            {
                tmp = kite_new_float(thd, *(double*)parameters[i]);
                break;
            }
            case 'p':
            {
                tmp = kite_new_integer(thd, *(int*)parameters[i]);
                break;
            }
            case 't':
            {
                tmp = kite_new_string(thd, *(char**)parameters[i]);
                break;
            }
            case '2':
            {
                tmp = kite_new_integer(thd, *(cinv_int16_t*)parameters[i]);
                break;
            }
            case '4':
            {
                tmp = kite_new_integer(thd, *(cinv_int32_t*)parameters[i]);
                break;
            }
            case '8':
            {
                tmp = kite_new_integer(thd, *(cinv_int64_t*)parameters[i]);
                break;
            }
            default:
            {
                assert(0);
            }
        }

        i++;
        kite_append_list(thd, methodArgs, tmp);
        param_types_string++;
    }
    
    // Call method
    curstack = thd->running_stack->length - 1;
    
    kite_funccall_info_t *entry = calloc(sizeof(kite_funccall_info_t), 1);
    entry->this = method->builtin_data.funcvalue.this;
    entry->file = "(C to Kite layer)";
    kite_push_stack(&thd->func_stack, FALSE, entry);

    if (!method->builtin_data.funcvalue.this) {
        kite_vm_call_object(thd, kite_new_null(thd), method, methodArgs);
    } else {
        kite_vm_call_object(thd, method->builtin_data.funcvalue.this, method, methodArgs);
    }
    if (thd->exception) return;

    /* Remove anything other than the return value from the top of the stack */
    kite_object_t *ret;
    if ((thd->running_stack->length - 1) != curstack) {
        ret = kite_vm_pop(thd);
        while((thd->running_stack->length - 1) != curstack) {
            if (thd->running_stack->stack[thd->running_stack->length - 1].reference) {
                (void)kite_vm_pop(thd);
            } else {
                tmp = kite_vm_pop(thd);
                kite_dereference_object(tmp);
            }
        }
    }
    POP_FUNC_STACK;
    
    ret_type_string = KITE_GET_STRING_VALUE(retString);
    switch(*ret_type_string)
    {
        case 'c':
        {
            *(char*)(returnout) = (char)KITE_GET_INTEGER(ret);
            break;
        }
        case 's':
        {
            *(short*)(returnout) = (short)KITE_GET_INTEGER(ret);
            break;
        }
        case 'i':
        {
            *(int*)(returnout) = (int)KITE_GET_INTEGER(ret);
            break;
        }
        case 'l':
        {
            *(long*)(returnout) = (long)KITE_GET_INTEGER(ret);
            break;
        }
        case 'p':
        {
            *(int*)(returnout) = (int)KITE_GET_INTEGER(ret);
            break;
        }
        case 't':
        {
            *(char**)(returnout) = KITE_GET_STRING_VALUE(ret);
            break;
        }
        case '2':
        {
            *(cinv_int16_t*)(returnout) = (cinv_int16_t)KITE_GET_INTEGER(ret);
            break;
        }
        case '4':
        {
            *(cinv_int32_t*)(returnout) = (cinv_int32_t)KITE_GET_INTEGER(ret);
            break;
        }
        case '8':
        {
            *(cinv_int64_t*)(returnout) = (cinv_int64_t)KITE_GET_INTEGER(ret);
            break;
        }
        case 'e':
        {
            *(long long*)(returnout) = (long long)KITE_GET_INTEGER(ret);
            break;
        }
        case 'f':
        {
            *(float*)(returnout) = (float)KITE_GET_FLOAT(ret);
            break;
        }
        case 'd':
        {
            *(double*)(returnout) = (double)KITE_GET_FLOAT(ret);
            break;
        }
        case 0:
        {
            // void
            break;
        }
        default:
        {
            assert(0);
        }
    }
    
    kite_dereference_object(methodArgs);
    return;
}

/*****************************************************************************
 * Class constructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeCallback_construct) {
    kite_object_t *ctxObj, *funcObj, *userData, *retString, *paramString;
    kite_object_t *dataList, *method;
    CInvContext *ctx;
    CInvFunction *func;
    
    (void)thd;
    
    KITE_GET_METHOD_ARGUMENT(ctxObj, 1);
    KITE_GET_METHOD_ARGUMENT(funcObj, 2);
    KITE_GET_METHOD_ARGUMENT(userData, 3);
    KITE_GET_METHOD_ARGUMENT(retString, 4);
    KITE_GET_METHOD_ARGUMENT(paramString, 5);
    KITE_GET_METHOD_ARGUMENT(method, 6);
    
    ctx = (CInvContext*)ctxObj->builtin_data.filevalue;
    func = (CInvFunction*)funcObj->builtin_data.listvalue.cdr;
    dataList = kite_new_list(thd);
    
    kite_append_list(thd, dataList, retString);
    kite_append_list(thd, dataList, paramString);
    kite_append_list(thd, dataList, userData);
    kite_append_list(thd, dataList, method);
    kite_append_list(thd, dataList, kite_new_integer(thd, (long)thd));

    this->builtin_data.listvalue.car = 
        (kite_object_t*)ctx;
    this->builtin_data.listvalue.cdr =
        (kite_object_t*)cinv_callback_create(
            ctx, 
            func,
            dataList, 
            CInvokeCallback_callIntoKite);
    this->builtin_data.listvalue.iterator = dataList;
}

/*****************************************************************************
 * Class destructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeCallback_destruct) {
    KITE_NO_ARGS;
    (void)thd;
    
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvCallback *cb = (CInvCallback*)this->builtin_data.listvalue.cdr;
    
    cinv_callback_delete(ctx, cb);
    kite_dereference_object(this->builtin_data.listvalue.iterator);
}

/*****************************************************************************
 * Get function entrypoint
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeCallback_entryPoint) {
    KITE_NO_ARGS;
    
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvCallback *cb = (CInvCallback*)this->builtin_data.listvalue.cdr;
    
    kite_vm_return(thd, 
        kite_new_integer(thd, 
            (long)cinv_callback_getentrypoint(
                ctx, 
                cb
            )
        )
    );
}
#endif

KITE_MODULE_INITIALIZER(_cinvoke_callback) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "_cinvoke_callback");
    kite_set_docstring(newclass, "CInvokeCallback internal code.");

#ifdef BUILD_CINVOKE    
    /* Add methods. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeCallback_construct, 
                                  "Class constructor.", 6,
                                  "ctx", "CInvokeContext instance",
                                  "func", "CInvokeFunction instance",
                                  "userData", "User data",
                                  "retString", "Return value type of callback.",
                                  "paramString", "Parameters of callback.",
                                  "method", "Kite method to call."));
    kite_add_method(thread, newclass, "destruct",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeCallback_destruct, 
                                  "Free associated memory and resources.", 0));
    kite_add_method(thread, newclass, "get_entrypoint",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeCallback_entryPoint, 
                                  "Retrieves function pointer for callback function.", 0));
#endif

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "_cinvoke_callback", 1, NULL);
    kite_set_property(thread, parent, "_cinvoke_callback", newclass);
}

==> kite_1.0.2.orig.tar.gz: kite-1.0.2/modules/__internal/_cinvoke_callback.c <==
/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "kite_object.h"
#include "vm/kite_vm.h"
#include "config.h"

#ifdef BUILD_CINVOKE
#include "cinvoke/cinvoke.h"

/*****************************************************************************
 * Arrange to call given Kite method.
 ****************************************************************************/
void
CInvokeCallback_callIntoKite(CInvFunction *f, void *parameters[], 
                             void *returnout, void *userdata)
{
    char *param_types_string, *ret_type_string;
    kite_object_t *args = (kite_object_t*)userdata;
    kite_object_t *retString, *paramString, *userData, *method;
    kite_object_t *methodArgs;
    kite_object_t *threadInt;
    kite_object_t *tmp;
    kite_thread_t *thd;
    int i = 0;
    int curstack;
    kite_object_t *ret = NULL;
    kite_funccall_info_t *fc_entry;

    (void)f;
    
    KITE_GET_METHOD_ARGUMENT(retString, 1);
    KITE_GET_METHOD_ARGUMENT(paramString, 2);
    KITE_GET_METHOD_ARGUMENT(userData, 3);
    KITE_GET_METHOD_ARGUMENT(method, 4);
    KITE_GET_METHOD_ARGUMENT(threadInt, 5);
    
    thd = (kite_thread_t*)KITE_GET_INTEGER(threadInt);
    param_types_string = KITE_GET_STRING_VALUE(paramString);
    methodArgs = kite_new_list(thd);
    kite_append_list(thd, methodArgs, userData);
    while(*param_types_string)
    {        
        switch(*param_types_string)
        {
            case 'c':
            {
                tmp = kite_new_integer(thd, *(char*)parameters[i]);
                break;
            }
            case 's':
            {
                tmp = kite_new_integer(thd, *(short*)parameters[i]);
                break;
            }
            case 'i':
            {
                tmp = kite_new_integer(thd, *(int*)parameters[i]);
                break;
            }
            case 'l':
            {
                tmp = kite_new_integer(thd, *(long*)parameters[i]);
                break;
            }
            case 'e':
            {
                tmp = kite_new_integer(thd, (long)*(long long*)parameters[i]);
                break;
            }
            case 'f':
            {
                tmp = kite_new_float(thd, *(float*)parameters[i]);
                break;
            }
            case 'd':
            {
                tmp = kite_new_float(thd, *(double*)parameters[i]);
                break;
            }
            case 'p':
            {
                tmp = kite_new_integer(thd, *(int*)parameters[i]);
                break;
            }
            case 't':
            {
                tmp = kite_new_string(thd, *(char**)parameters[i]);
                break;
            }
            case '2':
            {
                tmp = kite_new_integer(thd, *(cinv_int16_t*)parameters[i]);
                break;
            }
            case '4':
            {
                tmp = kite_new_integer(thd, *(cinv_int32_t*)parameters[i]);
                break;
            }
            case '8':
            {
                tmp = kite_new_integer(thd, (long)*(cinv_int64_t*)parameters[i]);
                break;
            }
            default:
            {
                assert(0);
            }
        }

        i++;
        kite_append_list(thd, methodArgs, tmp);
        param_types_string++;
    }
    
    // Call method
    curstack = thd->running_stack->length - 1;
    
    fc_entry = calloc(sizeof(kite_funccall_info_t), 1);
    fc_entry->this = method->builtin_data.funcvalue.this;
    fc_entry->file = "(C to Kite layer)";
    kite_push_stack(&thd->func_stack, FALSE, fc_entry);

    if (!method->builtin_data.funcvalue.this) {
        kite_vm_call_object(thd, kite_new_null(thd), method, methodArgs);
    } else {
        kite_vm_call_object(thd, method->builtin_data.funcvalue.this, method, methodArgs);
    }
    if (thd->exception) return;

    /* Remove anything other than the return value from the top of the stack */
    if ((thd->running_stack->length - 1) != curstack) {
        ret = kite_vm_pop(thd);
        while((thd->running_stack->length - 1) != curstack) {
            if (thd->running_stack->stack[thd->running_stack->length - 1].reference) {
                (void)kite_vm_pop(thd);
            } else {
                tmp = kite_vm_pop(thd);
                kite_dereference_object(tmp);
            }
        }
    }
    POP_FUNC_STACK;
    
    ret_type_string = KITE_GET_STRING_VALUE(retString);
    switch(*ret_type_string)
    {
        case 'c':
        {
            *(char*)(returnout) = (char)KITE_GET_INTEGER(ret);
            break;
        }
        case 's':
        {
            *(short*)(returnout) = (short)KITE_GET_INTEGER(ret);
            break;
        }
        case 'i':
        {
            *(int*)(returnout) = (int)KITE_GET_INTEGER(ret);
            break;
        }
        case 'l':
        {
            *(long*)(returnout) = (long)KITE_GET_INTEGER(ret);
            break;
        }
        case 'p':
        {
            *(int*)(returnout) = (int)KITE_GET_INTEGER(ret);
            break;
        }
        case 't':
        {
            *(char**)(returnout) = KITE_GET_STRING_VALUE(ret);
            break;
        }
        case '2':
        {
            *(cinv_int16_t*)(returnout) = (cinv_int16_t)KITE_GET_INTEGER(ret);
            break;
        }
        case '4':
        {
            *(cinv_int32_t*)(returnout) = (cinv_int32_t)KITE_GET_INTEGER(ret);
            break;
        }
        case '8':
        {
            *(cinv_int64_t*)(returnout) = (cinv_int64_t)KITE_GET_INTEGER(ret);
            break;
        }
        case 'e':
        {
            *(long long*)(returnout) = (long long)KITE_GET_INTEGER(ret);
            break;
        }
        case 'f':
        {
            *(float*)(returnout) = (float)KITE_GET_FLOAT(ret);
            break;
        }
        case 'd':
        {
            *(double*)(returnout) = (double)KITE_GET_FLOAT(ret);
            break;
        }
        case 0:
        {
            // void
            break;
        }
        default:
        {
            assert(0);
        }
    }
    
    kite_dereference_object(methodArgs);
    return;
}

/*****************************************************************************
 * Class constructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeCallback_construct) {
    kite_object_t *ctxObj, *funcObj, *userData, *retString, *paramString;
    kite_object_t *dataList, *method;
    CInvContext *ctx;
    CInvFunction *func;
    
    (void)thd;
    
    KITE_GET_METHOD_ARGUMENT(ctxObj, 1);
    KITE_GET_METHOD_ARGUMENT(funcObj, 2);
    KITE_GET_METHOD_ARGUMENT(userData, 3);
    KITE_GET_METHOD_ARGUMENT(retString, 4);
    KITE_GET_METHOD_ARGUMENT(paramString, 5);
    KITE_GET_METHOD_ARGUMENT(method, 6);
    
    ctx = (CInvContext*)ctxObj->builtin_data.filevalue;
    func = (CInvFunction*)funcObj->builtin_data.listvalue.cdr;
    dataList = kite_new_list(thd);
    
    kite_append_list(thd, dataList, retString);
    kite_append_list(thd, dataList, paramString);
    kite_append_list(thd, dataList, userData);
    kite_append_list(thd, dataList, method);
    kite_append_list(thd, dataList, kite_new_integer(thd, (long)thd));

    this->builtin_data.listvalue.car = 
        (kite_object_t*)ctx;
    this->builtin_data.listvalue.cdr =
        (kite_object_t*)cinv_callback_create(
            ctx, 
            func,
            dataList, 
            CInvokeCallback_callIntoKite);
    this->builtin_data.listvalue.iterator = dataList;
}

/*****************************************************************************
 * Class destructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeCallback_destruct) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvCallback *cb = (CInvCallback*)this->builtin_data.listvalue.cdr;
    
    KITE_NO_ARGS;
    (void)thd;
    
    cinv_callback_delete(ctx, cb);
    kite_dereference_object(this->builtin_data.listvalue.iterator);
}

/*****************************************************************************
 * Get function entrypoint
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeCallback_entryPoint) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvCallback *cb = (CInvCallback*)this->builtin_data.listvalue.cdr;
    
    KITE_NO_ARGS;
    
    kite_vm_return(thd, 
        kite_new_integer(thd, 
            (long)cinv_callback_getentrypoint(
                ctx, 
                cb
            )
        )
    );
}
#endif

KITE_MODULE_INITIALIZER(_cinvoke_callback) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "_cinvoke_callback");
    kite_set_docstring(newclass, "CInvokeCallback internal code.");

#ifdef BUILD_CINVOKE    
    /* Add methods. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method_compiled_with_docs(thread, CInvokeCallback_construct, 
                                  "Class constructor.", 6,
                                  "ctx", "CInvokeContext instance",
                                  "func", "CInvokeFunction instance",
                                  "userData", "User data",
                                  "retString", "Return value type of callback.",
                                  "paramString", "Parameters of callback.",
                                  "method", "Kite method to call."));
    kite_add_method(thread, newclass, "destruct",
        kite_new_method_compiled_with_docs(thread, CInvokeCallback_destruct, 
                                  "Free associated memory and resources.", 0));
    kite_add_method(thread, newclass, "get_entrypoint",
        kite_new_method_compiled_with_docs(thread, CInvokeCallback_entryPoint, 
                                  "Retrieves function pointer for callback function.", 0));
#endif

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "_cinvoke_callback", 1, NULL);
    kite_set_property(thread, parent, "_cinvoke_callback", newclass);
}
